%
%
%  Filaments Create Channel Function for Imaris 7.3.0
%
%  Copyright Bitplane AG 2011
%
%
%  Installation:
%  - Copy this file into the XTensions folder in the Imaris installation directory.
%  - You will find this function in the Image Processing menu
%
%    <CustomTools>
%      <Menu>
%       <Submenu name="Filaments Functions">
%        <Item name="Create Channel from Filaments" icon="Matlab">
%          <Command>MatlabXT::XTFilamentsCreateChannel(%i)</Command>
%        </Item>
%       </Submenu>
%      </Menu>
%      <SurpassTab>
%        <SurpassComponent name="bpFilaments">
%          <Item name="Create Channel" icon="Matlab">
%            <Command>MatlabXT::XTFilamentsCreateChannel(%i)</Command>
%          </Item>
%        </SurpassComponent>
%      </SurpassTab>
%    </CustomTools>
% 
% !!! Additional informations:
% A new vChannel is added to the dataset
% Original dataset size is maintained
%

function XTFilamentsCreateChannel(aImarisApplicationID, aDataSetResolution)

% connect to Imaris interface
if ~isa(aImarisApplicationID, 'Imaris.IApplicationPrxHelper')
  javaaddpath ImarisLib.jar
  vImarisLib = ImarisLib;
  if ischar(aImarisApplicationID)
    aImarisApplicationID = round(str2double(aImarisApplicationID));
  end
  vImarisApplication = vImarisLib.GetApplication(aImarisApplicationID);
else
  vImarisApplication = aImarisApplicationID;
end

% the user has to create a scene with some filament
vSurpassScene = vImarisApplication.GetSurpassScene;
if isequal(vSurpassScene, [])
    msgbox('Please create some Filaments in the Surpass scene!');
    return;
end
vImarisApplication.DataSetPushUndo('Create Channel from Filaments');

% parameter
% aDataSetResolution: 
%   if it is > 0, 
%     it is the number of pixels (+/- cubic) along princial dimension
%     of a NEW dataset
%   if it is = 0, 
%     a new vChannel is added to the Imaris dataset (size is unchanged)
if nargin < 2
    aDataSetResolution = 0;
end

% get the filament
vFactory = vImarisApplication.GetFactory;
vFilaments = vFactory.ToFilaments(vImarisApplication.GetSurpassSelection);

% search the filament if not previously selected
vSurpassScene = vImarisApplication.GetSurpassScene;
if ~vFactory.IsFilaments(vFilaments)        
    for vChildIndex = 1:vSurpassScene.GetNumberOfChildren
        vDataItem = vSurpassScene.GetChild(vChildIndex - 1);
        if vFactory.IsFilaments(vDataItem)
            vFilaments = vFactory.ToFilaments(vDataItem);
            break;
        end
    end
    % did we find the filament?
    if isequal(vFilaments, [])
        msgbox('Please create some filament!');
        return;
    end
end

vInterpolate = questdlg('Interpolate intensity along filament diameter?', ...
  'XTFilamentCreateChannel', 'Yes', 'No', 'Cancel', 'Yes');
if isequal(vInterpolate, 'Cancel'), return, end
vInterpolate = isequal(vInterpolate, 'Yes');

if aDataSetResolution > 0
    % create new dataset with this resolution
    vExtends = [0, 0, 0; 100, 100, 10];
    [vDataSet, vData, vMin, vMax, vType] = InitNewDataSet(vImarisApplication, ...
        vExtends, aDataSetResolution);
else
    % add vChannel to current dataset
    [vDataSet, vData, vMin, vMax, vType] = InitNewChannel(vImarisApplication);
end
vChannel = vDataSet.GetSizeC - 1;

vWaitBar = waitbar(0, 'MaskFilaments');
vCount = vFilaments.GetNumberOfFilaments;

% sort filaments by time point, then process one time point after the other
vTimes = 1:vCount;
for vIndex = 1:vCount
  vTimes(vIndex) = vFilaments.GetTimeIndex(vIndex-1);
end
[vTimes, vOrder] = sort(vTimes);

for vIndexF = 1:vCount
  vFilamentIndex = vOrder(vIndexF) - 1;
  vPos = vFilaments.GetPositionsXYZ(vFilamentIndex);
  vEdges = vFilaments.GetEdges(vFilamentIndex) + 1;
  vRadii = vFilaments.GetRadii(vFilamentIndex);
    
  % vTerminals = 1:size(vPos, 1);
  vTerminals = find(FindTerminals(size(vPos, 1), vEdges(:)));

  vNumTerminals = numel(vTerminals);
  vSteps = vNumTerminals + size(vEdges, 1);
  for vIndex = 1:vNumTerminals
      vPosition = vPos(vTerminals(vIndex), :);
      vRadius = vRadii(vTerminals(vIndex));
      % copy overhead: 5.1 -> 5.67, 10.2 -> 10.54
      vData = DrawSphere(vData, vPosition, vRadius, vMin, vMax, vType, vInterpolate);

      if mod(vIndex, 20) == 0
          waitbar((vIndexF-1+vIndex/vSteps)/vCount, vWaitBar)
      end
  end

  for vIndex = 1:size(vEdges, 1)
      vPosition1 = vPos(vEdges(vIndex, 1), :);
      vPosition2 = vPos(vEdges(vIndex, 2), :);
      vRadius1 = vRadii(vEdges(vIndex, 1));
      vRadius2 = vRadii(vEdges(vIndex, 2));
      vData = DrawDisk(vData, vPosition1, vPosition2, vRadius1, vRadius2, ...
        vMin, vMax, vType, vInterpolate);
      if mod(vIndex, 20) == 0
          waitbar((vIndexF-1+(vNumTerminals+vIndex)/vSteps)/vCount, vWaitBar)
      end
  end 

  vTime = vFilaments.GetTimeIndex(vFilamentIndex);
  if vIndexF == numel(vOrder) || ...
      vTime ~= vFilaments.GetTimeIndex(vOrder(vIndexF+1)-1)
    if strcmp(vDataSet.GetType, 'eTypeUInt8')
      for vIndexZ = 1:size(vData, 3)
        vDataSet.SetDataSliceBytes(vData(:, :, vIndexZ), vIndexZ-1, vChannel, vTime);
      end
    elseif strcmp(vDataSet.GetType, 'eTypeUInt16')
      for vIndexZ = 1:size(vData, 3)
        vDataSet.SetDataSliceShorts(vData(:, :, vIndexZ), vIndexZ-1, vChannel, vTime);
      end
    else
      for vIndexZ = 1:size(vData, 3)
        vDataSet.SetDataSliceFloats(vData(:, :, vIndexZ), vIndexZ-1, vChannel, vTime);
      end
    end
    if vIndexF < numel(vOrder)
      vData(:) = 0;
    end
  end
  
  waitbar(vIndexF/vCount, vWaitBar)
end
close(vWaitBar)

vRGBA = vFilaments.GetColorRGBA;
vDataSet.SetChannelColorRGBA(vChannel, vRGBA);
vDataSet.SetChannelName(vChannel, char(vFilaments.GetName));
vImarisApplication.SetDataSet(vDataSet);


function aData = DrawSphere(aData, aPos, aRad, aMin, aMax, aType, aInterpolate)
vSize = [size(aData, 1), size(aData, 2), size(aData, 3)];
aPos = (aPos - aMin) ./ (aMax - aMin) .* vSize + 0.5;
aRad = aRad ./ (aMax - aMin) .* vSize;
vPosMin = round(max(aPos - aRad, 1));
vPosMax = round(min(aPos + aRad, vSize));
vPosX = vPosMin(1):vPosMax(1);
vPosY = vPosMin(2):vPosMax(2);
vPosZ = vPosMin(3):vPosMax(3);
[vX, vY, vZ] = ndgrid(vPosX, vPosY, vPosZ);
vDist = ((vX - aPos(1))/aRad(1)).^2 + ((vY - aPos(2))/aRad(2)).^2 + ...
     ((vZ - aPos(3))/aRad(3)).^2;
vInside = vDist < 1;
vCube = aData(vPosX, vPosY, vPosZ);
if aInterpolate
  vCube(vInside) = max(vCube(vInside), FixType(255*(1-vDist(vInside)), aType));
else
  vCube(vInside) = 255;
end
aData(vPosX, vPosY, vPosZ) = vCube;


function aData = DrawDisk(aData, aPos1, aPos2, aRad1, aRad2, aMin, aMax, ...
  aType, aInterpolate)
vSize = [size(aData, 1), size(aData, 2), size(aData, 3)];
aPos1 = (aPos1 - aMin) ./ (aMax - aMin) .* vSize + 0.5;
aPos2 = (aPos2 - aMin) ./ (aMax - aMin) .* vSize + 0.5;
aRad1 = aRad1 ./ (aMax - aMin) .* vSize;
aRad2 = aRad2 ./ (aMax - aMin) .* vSize;
vPosMin = round(max(min(aPos1 - aRad1, aPos2 - aRad2), 1));
vPosMax = round(min(max(aPos1 + aRad1, aPos2 + aRad2), vSize));
vPosX = vPosMin(1):vPosMax(1);
vPosY = vPosMin(2):vPosMax(2);
vPosZ = vPosMin(3):vPosMax(3);
[vX, vY, vZ] = ndgrid(vPosX, vPosY, vPosZ);
vAxis = aPos2 - aPos1;
vNorm = norm(vAxis);
vAxis = vAxis / vNorm;
vx = vX - aPos1(1);
vy = vY - aPos1(2);
vz = vZ - aPos1(3);
uv = vx*vAxis(1) + vy*vAxis(2) + vz*vAxis(3);
vx = aPos1(1) + uv * vAxis(1);
vy = aPos1(2) + uv * vAxis(2);
vz = aPos1(3) + uv * vAxis(3);
radx = aRad1(1) + (aRad2(1) - aRad1(1)) * uv / vNorm;
rady = aRad1(2) + (aRad2(2) - aRad1(2)) * uv / vNorm;
radz = aRad1(3) + (aRad2(3) - aRad1(3)) * uv / vNorm;
intern = uv >= 0 & uv <= vNorm;
vDist = ((vX - vx)./radx).^2 + ((vY - vy)./rady).^2 + ((vZ - vz)./radz).^2;
vInside = intern & vDist < 1;
vCube = aData(vPosX, vPosY, vPosZ);
if aInterpolate
  vCube(vInside) = max(vCube(vInside), FixType(255*(1-vDist(vInside)), aType));
else
  vCube(vInside) = 255;
end
aData(vPosX, vPosY, vPosZ) = vCube;


function aData = FixType(aData, aType)
if strcmp(aType, 'eTypeUInt8')
  aData = uint8(aData);
elseif strcmp(aType, 'eTypeUInt16')
  aData = uint16(aData);
else
  aData = single(aData);
end


function aIsTerminal = FindTerminals(aNumPoints, aEdgesElements)
aIsTerminal = true(1, aNumPoints);
vIndex(aEdgesElements) = 1:numel(aEdgesElements);
vIndex = vIndex(vIndex > 0); % holes :(
aEdgesElements(vIndex) = 0; % aEdgesElements(vIndex) = []; is slower :/
aEdgesElements = aEdgesElements(aEdgesElements > 0);
aIsTerminal(aEdgesElements) = false;


function [aDataSet, aData, aMin, aMax, aType] = InitNewDataSet(aImaris, ...
    aPos, aDataSetResolution)
aMin = min(aPos, [], 1);
aMax = max(aPos, [], 1);
vSize = aMax - aMin;
vSize = round(vSize / max(vSize) * aDataSetResolution); 
aData = zeros(vSize, 'uint8');
aDataSet = aImaris.GetFactory.CreateDataSet;
aDataSet.Create('eTypeUInt8', vSize(1), vSize(2), vSize(3));
aDataSet.SetExtendMinX(aMin(1));
aDataSet.SetExtendMinY(aMin(2));
aDataSet.SetExtendMinZ(aMin(3));
aDataSet.SetExtendMaxX(aMax(1));
aDataSet.SetExtendMaxY(aMax(2));
aDataSet.SetExtendMaxZ(aMax(3));
aType = aDataSet.GetType;


function [aDataSet, aData, aMin, aMax, aType] = InitNewChannel(aImaris)
aDataSet = aImaris.GetDataSet.Clone;
aMin = [aDataSet.GetExtendMinX, aDataSet.GetExtendMinY, aDataSet.GetExtendMinZ];
aMax = [aDataSet.GetExtendMaxX, aDataSet.GetExtendMaxY, aDataSet.GetExtendMaxZ];
vSize = [aDataSet.GetSizeX, aDataSet.GetSizeY, aDataSet.GetSizeZ];
if strcmp(aDataSet.GetType, 'eTypeUInt8')
  aData = zeros(vSize, 'uint8');
elseif strcmp(aDataSet.GetType, 'eTypeUInt16')
  aData = zeros(vSize, 'uint16');
else
  aData = zeros(vSize, 'single');
end
aDataSet.SetSizeC(aDataSet.GetSizeC + 1);
aType = aDataSet.GetType;
